# -*- coding: utf-8 -*-
"""
Очередь на базе RabbitMQ.
"""
from __future__ import print_function
from __future__ import absolute_import
from pushme.mq import interface
import pika
from pika.adapters.tornado_connection import TornadoConnection

EXCHANGE_NAME = 'pushme-exchange'
QUEUE_NAME = 'pushme'


def create_url_parameters(host, port, login, password, vhost):
    url = 'amqp://{login}:{password}@{host}:{port}/{vhost}'.format(
        login=login or 'guest',
        password=password or 'guest',
        host=host or 'localhost',
        port=port or 5672,
        vhost=vhost,
    )

    return pika.URLParameters(url)


class Queue(interface.Queue):
    """
    Заглушка для процесса Queue, так как очередь реализует сервер RabbitMQ.
    """

    def run(self):
        print("There is no need to start the Queue process  "
              "when using the RabbitMQ backend.")


class Client(interface._Client):
    """
    Базовый класс клиента очереди
    """
    def _create_parameters(self):
        return create_url_parameters(
            *self._queue_addr,
            *self._queue_access,
        )


class Sender(interface.Sender, Client):
    """
    Отправитель
    """
    def __init__(self, *args, **kwargs):
        super(Sender, self).__init__(*args, **kwargs)
        self._connection = pika.BlockingConnection(self._create_parameters())
        self._channel = self._connection.channel()
        self._channel.exchange_declare(
            exchange=EXCHANGE_NAME, exchange_type='fanout'
        )

    def send(self, data, uid=None, topic=None):
        message = (
            '%s|%s|%s' % (
                uid or 0,
                topic or "",
                data
        )).encode()
        self._channel.basic_publish(
            EXCHANGE_NAME, '', message,
            pika.BasicProperties(
                content_type='text/plain', delivery_mode=1)
        )


class Receiver(interface.Receiver, Client):
    """
    Получатель
    """
    def __init__(self, *args, **kwargs):
        super(Receiver, self).__init__(*args, **kwargs)
        self._channel = None
        self._connection = self.connect()

    def connect(self):
        """Устанавливает соединение с RabbitMQ сервером"""
        return TornadoConnection(
            self._create_parameters(),
            self.on_connected,
            on_open_error_callback=self.on_connection_closed,
        )

    def reconnect(self):
        """Переустанавливает закрытое соединение с сервером"""
        self._connection = self.connect()

    def on_connected(self, connection):
        """Вызывается, когда выполнено подключение к RabbitMQ"""
        self._connection.add_on_close_callback(self.on_connection_closed)
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_connection_closed(self, connection, reason):
        """Вызывается, когда подключение закрывается и при ошибке соединения"""
        # попытаемся заново устакновить соединение через 5 секунд
        self._connection.ioloop.call_later(5, self.reconnect)

    def on_channel_open(self, new_channel):
        """Вызывается, когда открыт канал"""
        self._channel = new_channel
        self._channel.exchange_declare(EXCHANGE_NAME, exchange_type='fanout',
                                       callback=self.on_exchange_declare)

    def on_exchange_declare(self, unused_frame):
        """Вызывается после успешного создания exchange"""
        self._channel.queue_declare(
            queue=QUEUE_NAME, durable=True, exclusive=False,
            auto_delete=False, callback=self.on_open_queue)

    def on_open_queue(self, channel):
        """Вызывается после успешного открытия очереди"""
        self._channel.queue_bind(QUEUE_NAME, EXCHANGE_NAME,
                                 callback=self.on_bind)

    def on_bind(self, unused_frame):
        """Вызывается после получения Queue.BindOk при вызове queue_bind()"""
        self._channel.basic_consume(queue=QUEUE_NAME, auto_ack=True,
                                    on_message_callback=self.handle_delivery)

    def handle_delivery(self, channel, method, header, body):
        """Вызывается, когда пришло сообщение из очереди"""
        try:
            message = body.decode('utf-8')
            uid, topic, data = message.split('|', 2)
            topic = topic.strip() or None
            uid = int(uid)
        except (ValueError, TypeError):
            print("Wrong message format:", body)
        else:
            self.callback(uid, topic, data)

    def process(self):
        # Сделаем заглушку, так как работаем через TornadoAdapter, а у него
        # свой loop
        pass
